/*

    zoltan_core.c - ZoltanPlayer CD/MP3 Player

    Core functions.

    Copyright (C) 2001 Angel Ortega <angel@triptico.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    http://www.triptico.com

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <dirent.h>
#include <linux/cdrom.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <linux/soundcard.h>
#include <time.h>

#include "zoltan.h"


/*******************
	Data
********************/

/* the songs */
struct song songs[MAX_SONGS];

/* The groups */
struct group groups[MAX_GROUPS];

/**
 * _num_songs - Number of songs in current media
 *
 * Contains the total number of songs in current media.
 */
int _num_songs;

/**
 * _num_groups - Number of groups in current media
 *
 * Contains the total number of groups in current media.
 * Should be 1 for Audio CDs.
 */
int _num_groups;

/* the CD file descriptor */
int _cd_fd=-1;

/* the mixer file descriptor */
int _mixer_fd=-1;

/* media type */
int _media_type=TYPE_UNKNOWN;

/* CD Audio data */
int _cda_total_time=0;
int _cda_num_tracks=0;

/* temporary buffers */
static char buffer[8192];

/* external player pid */
int _player_pid=-1;

/* cddb id */
int _cddb_id=0;

/* pause flag */
int _zoltan_paused=0;

/* song startup time */
int _song_startup_time=0;

/**
 * _current_song - The song currently being played
 *
 * Contains the song currently being played, or ready for it.
 * A value of -1 means no song (i.e., media is stopped).
 */
int _current_song;


/* configuration */

/* cd device */
char _cd_device[128]="/dev/cdrom";

/* mixer device */
char _mixer_device[128]="";

/* mount point */
char _mount_point[128]="/cdrom";

/* songs on hd */
char _hd_song_dir[128]="";

/* paths to players */

struct player
{
	char * ext;
	char * path;
} players[MAX_PLAYERS];

int _num_players=0;

/* path to cddb cache */
char _cddb_cache[128]="./cddb";

/* the TCP/IP port */
int _zoltan_port=6543;

/* volume (int percent) */
int _zoltan_volume=80;

/* volume increment/decrement with each command */
int _zoltan_volume_increment=5;

/* command to exec when the song has changed */
char _notify_command[2048]="";


/* digital songs cache */

#define SONG_CACHE_FORMAT	1

int song_cache[SONG_CACHE_SIZE];

int _song_cache_updates=0;

static char _song_cache_file[128]="/var/cache/zoltan.dat";

/* shuffle data */

int _zoltan_shuffle=0;

static int _song_order[MAX_SONGS];


/*******************
	Code
********************/

/**
 * _zoltan_sigchld - Signal trap for player death
 * @s: the signal
 *
 * Signal trap for player (child) death.
 */
static void _zoltan_sigchld(int s)
{
	int status;

/*	  pid=wait(&status);

	if(pid==_player_pid)
		_player_pid=-1; */

	if(waitpid(_player_pid,&status,WNOHANG)!=0)
	{
		_player_pid=-1;

		/* store current running time */
		if(_song_startup_time && songs[_current_song].filename!=NULL)
		{
			songs[_current_song].length=
				(int)time(NULL) - _song_startup_time;

			zoltan_cache_store_song();

			_song_startup_time=0;
		}
	}
}


/**
 * zoltan_cd_test - CD availability test
 *
 * Returns true if there is a CDA / CDROM in drive.
 */
int zoltan_cd_test(void)
{
	struct cdrom_subchnl cdsc;

	/* close tray, if open */
	ioctl(_cd_fd, CDROMCLOSETRAY);

	/* read the subchannel info */
	memset(&cdsc,'\0',sizeof(struct cdrom_subchnl));
	cdsc.cdsc_format=CDROM_MSF;

	if(ioctl(_cd_fd, CDROMSUBCHNL, &cdsc) < 0)
		return(0);

	return(1);
}


/**
 * _zoltan_cddb_id - Calculates the CDDB id
 *
 * Calculates the CDDB id of the inserted CD Audio.
 */
static unsigned int _zoltan_cddb_id(void)
{
	int n;
	unsigned int t,i;
	unsigned int cddb;
	char tmp[32];
	char * ptr;

	for(t=n=0;n < _cda_num_tracks;n++)
	{
		snprintf(tmp,sizeof(tmp),"%lu", (unsigned long int) songs[n].pos);

		for(ptr=tmp;*ptr!='\0';ptr++)
			t+=(*ptr - '0');
	}

	i=_cda_total_time - songs[0].pos;

	cddb=(t % 0xff) << 24 | i << 8 | _cda_num_tracks;

	printf("%08x\n",cddb);

	return(cddb);
}


/**
 * zoltan_cda_info - Gets info about the CD Audio
 *
 * Gets information about the CD Audio, setting the global
 * variables _cda_num_tracks, _cda_total_time and _cddb_id
 * and calculating the length (in seconds) of the songs.
 * Returns zero if some error ocurred.
 */
int zoltan_cda_info(void)
{
	struct cdrom_tochdr cdth;
	struct cdrom_tocentry cdte;
	int n;

	/* read the toc header */
	memset(&cdth,'\0',sizeof(struct cdrom_tochdr));

	if(ioctl(_cd_fd, CDROMREADTOCHDR, &cdth) < 0)
		return(0);

	_cda_num_tracks=cdth.cdth_trk1;

	/* read the tracks */
	for(n=0;n <= _cda_num_tracks && n < MAX_SONGS;n++)
	{
		cdte.cdte_format=CDROM_MSF;
		cdte.cdte_track=(n==_cda_num_tracks ? CDROM_LEADOUT : n+1);

		ioctl(_cd_fd, CDROMREADTOCENTRY, &cdte);

		songs[n].pos=(cdte.cdte_addr.msf.minute * 60) +
			cdte.cdte_addr.msf.second;

		if(n>0) songs[n-1].length=songs[n].pos - songs[n-1].pos;
	}

	_cda_total_time=songs[_cda_num_tracks].pos;

	_cddb_id=_zoltan_cddb_id();

	return(1);
}


#define SET_SONG_DATA(d) if(d!=NULL) { \
			 s->d=(char *)malloc(strlen(d)+1); \
			 strcpy(s->d,d); }

/**
 * zoltan_add_song - Adds a song to the internal database
 * @songinfo: text string about the song, human readable
 * @filename: the name of the file that stores the song, if applicable
 * @path: the path where the song file lives, if applicable
 * @group: group of the song
 * @player: the player to be used when playing the song
 *
 * Adds a song to the internal database. The global variable
 * _num_songs is incremented. All parameters are optional.
 */
void zoltan_add_song(char * songinfo, char * filename,
		     char * path, int group, int player)
{
	struct song * s;
	char tmp[1024];

	printf("[%s][%s][%s]\n",songinfo,filename,path);

	/* this should not occur, but who knows */
	if(_num_songs==MAX_SONGS) return;

	s=&songs[_num_songs];

	if(s->songinfo) { free(s->songinfo); s->songinfo=NULL; }
	if(s->filename) { free(s->filename); s->filename=NULL; }
	if(s->path) { free(s->path); s->path=NULL; }

	SET_SONG_DATA(songinfo);
	SET_SONG_DATA(filename);
	SET_SONG_DATA(path);

	s->songid=_num_songs;
	s->group=group;
	s->player=player;
	s->skip=0;

	if(filename!=NULL)
	{
		snprintf(tmp,sizeof(tmp),"%s/%s",path,filename);
		s->length=zoltan_search_cache(tmp);
	}

	_num_songs++;
}


/**
 * zoltan_add_group - Adds a new group to the internal database
 * @groupinfo: text string about the group, human readable
 * @is_playlist: 1 if the group is a playlist
 *
 * Adds a new group to the internal database. All parameters
 * are optional. The global variable _num_groups is incremented.
 */
void zoltan_add_group(char * groupinfo, int is_playlist)
{
	struct group * s;

	/* this should not occur, but who knows */
	if(_num_groups==MAX_GROUPS) return;

	s=&groups[_num_groups];

	if(s->groupinfo) { free(s->groupinfo); s->groupinfo=NULL; }

	memset(s,'\0',sizeof(struct group));

	SET_SONG_DATA(groupinfo);
	s->is_playlist=is_playlist;

	_num_groups++;

	printf("Addgroup [%s][%d]\n", groupinfo, _num_groups);
}


void zoltan_try_song(char * filename, char * directory, int group)
{
	char buf[2048];
	char * ext;
	int n;

	strncpy(buf, filename, sizeof(buf));

	/* regular file; test if playable */
	if((ext=strrchr(buf, '.'))!=NULL)
	{
		for(n=0;n < _num_players;n++)
		{
			if(strcasecmp(ext,players[n].ext)==0)
			{
				*ext='\0';

				zoltan_add_song(buf, filename,
					directory, group, n);

				break;
			}
		}
	}
}


void zoltan_parse_playlist(char * root, char * m3u, int group)
{
	char line[1024];
	FILE * f;
	char * ptr;

	printf("Playlist [%s]\n",m3u);

	if((f=fopen(m3u, "r"))==NULL)
		return;

	strncpy(line,m3u,sizeof(line));

	if((ptr=strrchr(line,'.'))!=NULL)
		*ptr='\0';

	zoltan_add_group(line + strlen(root) + 1, 1);

	while(fgets(line,sizeof(line),f)!=NULL)
	{
		line[strlen(line)-1]='\0';
		if(line[strlen(line)-1]=='\r')
			line[strlen(line)-1]='\0';

		if(line[0]=='#') continue;

		zoltan_try_song(line, root, group);
	}

	fclose(f);
}


/**
 * zoltan_song_find - Travel a directory tree searching for songs
 * @directory: the directory
 *
 * Searches (recursively) a directory, searching for playable
 * songs. If any is found, it's stored in the database. The
 * directories are stored as groups.
 */
void zoltan_song_find(char * root, char * directory, int group)
{
	DIR * d;
	struct dirent * de;
	struct stat s;
	char * ext;
	char newbase[1024];

	zoltan_add_group(directory + strlen(root) + 1, 0);

	d=opendir(directory);

	/* process plain files first */
	while((de=readdir(d))!=NULL)
	{
		if(strcmp(de->d_name,".")==0 ||
		   strcmp(de->d_name,"..")==0)
			continue;

		snprintf(newbase,sizeof(newbase),
			"%s/%s",directory,de->d_name);

		/* get file info */
		if(stat(newbase, &s) < 0)
			continue;

		if(S_ISREG(s.st_mode))
			zoltan_try_song(de->d_name, directory, group);
	}

	rewinddir(d);

	/* process subdirs and playlists now */
	while((de=readdir(d))!=NULL)
	{
		if(strcmp(de->d_name,".")==0 ||
		   strcmp(de->d_name,"..")==0)
			continue;

		snprintf(newbase,sizeof(newbase),
			"%s/%s",directory,de->d_name);

		/* get file info */
		if(stat(newbase, &s) < 0)
			continue;

		strncpy(buffer, de->d_name, sizeof(buffer));

		if(S_ISDIR(s.st_mode))
			zoltan_song_find(root, newbase, _num_groups);
		else
		if((ext=strrchr(de->d_name,'.'))!=NULL &&
		   strcasecmp(ext,".m3u")==0)
			zoltan_parse_playlist(directory, newbase, _num_groups);
	}

	closedir(d);
}


/**
 * _zoltan_kill_player - Kills the external player, if any
 *
 * Kills the external player, if it's running.
 */
static void _zoltan_kill_player(void)
{
	/* kill the running player, if any */
	if(_player_pid!=-1)
	{
		/* erase song startup time */
		_song_startup_time=0;

		/* die, please */
		kill(_player_pid, SIGTERM);

		usleep(100000);

		if(_player_pid!=-1)
		{
			/* die, bastard */
			kill(_player_pid, SIGKILL);
		}
	}

	_zoltan_paused=0;
}


/**
 * zoltan_notify - Notifies change of song
 *
 * If current song is different from the previously one
 * being played, notify using the external program.
 */
void zoltan_notify(void)
{
	static int _song=-1;
	char tmp[8192];

	if(_media_type==TYPE_UNKNOWN)
		return;

	if(_notify_command[0]!='\0' && _song != _current_song)
	{
		/* song has changed and we must inform about it */
		snprintf(tmp,sizeof(tmp),_notify_command,zoltan_build_song_info());
		system(tmp);
	}

	_song=_current_song;
}


/**
 * zoltan_spawn_player - Spawns an external player
 *
 * Strips path from filename, changes to that directory,
 * parses the external player exec line (substituting possible
 * %s strings with the song name) and spawns the external player.
 * Only returns to caller on player exec error.
 */
static void zoltan_spawn_player(void)
{
	char tmp[4096];
	char * argv[100];
	int n,u=0;
	char * ptr;
	char * path;

	/* change to directory */
	if(songs[_current_song].path!=NULL)
		chdir(songs[_current_song].path);

	/* strip command line */
	strncpy(tmp,players[songs[_current_song].player].path,sizeof(tmp));

	path=strtok(tmp," ");
	argv[0]="zoltan_player";

	for(n=1;(ptr=strtok(NULL," "));n++)
	{
		if(strcmp(ptr,"%s")==0)
		{
			u=1;
			argv[n]=songs[_current_song].filename;
		}
		else
			argv[n]=ptr;
	}

	if(!u) argv[n++]=songs[_current_song].filename;
	argv[n]=NULL;

	execv(path,argv);
/*
	execl(players[songs[_current_song].player].path,
		"zoltan_player",
		songs[_current_song].filename,NULL);
*/

	perror("Error: can't spawn external player");
}


/**
 * zoltan_play - Plays current song
 *
 * Plays the currently selected song.
 */
void zoltan_play(void)
{
	struct cdrom_ti cdti;

	if(_media_type==TYPE_CDA)
	{
		ioctl(_cd_fd, CDROMSTOP);

		memset(&cdti,'\0',sizeof(struct cdrom_ti));
		cdti.cdti_trk0=_current_song+1;
		cdti.cdti_trk1=_num_songs;

		ioctl(_cd_fd, CDROMPLAYTRKIND, &cdti);

		/* record startup time */
		_song_startup_time=(int)time(NULL);
	}
	else
	{
		_zoltan_kill_player();

		/* trap the signal */
		signal(SIGCHLD, _zoltan_sigchld);

		/* record startup time */
		_song_startup_time=(int)time(NULL);

		/* spawn new player */
		if((_player_pid=fork())==0)
		{
			/* this is the child */

			close(5);
			zoltan_spawn_player();
		}

		printf("New player is %d\n",_player_pid);
	}

	zoltan_notify();
}


/**
 * zoltan_stop - Stops the currently playing song
 *
 * Stops the currently playing song, if any.
 */
void zoltan_stop(void)
{
	if(_media_type==TYPE_CDA)
	{
		ioctl(_cd_fd, CDROMSTOP);
		_song_startup_time=0;
	}
	else
	if(_media_type==TYPE_CDROM || _media_type==TYPE_HD)
	{
		_zoltan_kill_player();
	}

	_current_song=-1;
	zoltan_notify();
}


/**
 * zoltan_pause - (Un)pauses the currently playing song
 *
 * Unpauses the currently playing song.
 */
void zoltan_pause(void)
{
	if(_zoltan_paused)
	{
		if(_media_type==TYPE_CDA)
			ioctl(_cd_fd, CDROMRESUME);
		else
		{
			if(_player_pid != -1)
				kill(_player_pid, SIGCONT);
		}

		_zoltan_paused=0;
	}
	else
	{
		if(_media_type==TYPE_CDA)
			ioctl(_cd_fd, CDROMPAUSE);
		else
		{
			/* SIGCHLD handler must be disabled, because
			   the child being stopped generates
			   a SIGCHLD to inform daddy of the change;
			   if we just leave like this, we'll
			   wait in an infinite wait() */

			if(_player_pid != -1)
			{
				signal(SIGCHLD, SIG_IGN);

				kill(_player_pid, SIGSTOP);

				/* restore */
				signal(SIGCHLD, _zoltan_sigchld);
			}
		}

		_zoltan_paused=1;
	}
}


void zoltan_change_song(int inc)
{
	int n;

	/* find current position */
	if(_current_song == -1)
	{
		n= inc > 0 ? -1 : _num_songs;
	}
	else
	{
		for(n=0;n < _num_songs;n++)
		{
			if(_song_order[n]==_current_song)
				break;
		}
	}

	if(inc > 0)
	{
		/* move to next */
		while(n < _num_songs)
		{
			n++;

			if(!songs[_song_order[n]].skip)
				break;
		}

		if(n == _num_songs)
			_current_song=-1;
		else
			_current_song=_song_order[n];
	}
	else
	{
		/* move to previous */
		while(n >= 0)
		{
			n--;

			if(!songs[_song_order[n]].skip)
				break;
		}

		if(n == -1)
			_current_song=-1;
		else
			_current_song=_song_order[n];
	}
}


/**
 * zoltan_change_group - Change the group of songs
 * @inc: number of groups to jump (positive forward, negative backward)
 *
 * Change the group of songs adding the increment to the group of
 * the current song. A call to zoltan_play() should follow.
 * A CD Audio has only one group, so any group change makes it
 * restart from the beginning.
 */
void zoltan_change_group(int inc)
{
	int c;

	c=songs[_current_song].group + inc;

	if(inc < 0)
	{
		if(c < 0)
			c=_num_groups - 1 + inc;

		/* search backwards */
		for(_current_song=_num_songs - 1;_current_song> - 1;_current_song--)
		{
			if(songs[_current_song].group < c - 1)
				break;
		}

		_current_song++;
	}
	else
	{
		if(c >= _num_groups)
			c=0;

		/* search the first song of the available group */
		for(_current_song=0;_current_song < _num_songs;_current_song++)
		{
			if(songs[_current_song].group >= c)
				break;
		}
	}

	zoltan_notify();
}


/**
 * zoltan_build_song_info - Builds a string with the song info
 *
 * Builds a string with the song information (song name, song
 * group and time duration, where available). Returns a pointer
 * to a static buffer.
 */
char * zoltan_build_song_info(void)
{
	static char songinfo[4096];
	char tmp[32];
	char * ptr;

	if(_current_song==-1)
	{
		strcpy(songinfo, "Idle");
		return(songinfo);
	}

	/* caution: no buffer overflowing is tested */

	songinfo[0]='\0';

	if((ptr=songs[_current_song].songinfo)!=NULL)
		strcpy(songinfo,ptr);

	if((ptr=groups[songs[_current_song].group].groupinfo)!=NULL && *ptr!='\0')
	{
		strcat(songinfo," (");
		strcat(songinfo,ptr);
		strcat(songinfo,")");
	}

	if(songs[_current_song].length)
	{
		sprintf(tmp," [%02d:%02d]",
			songs[_current_song].length / 60,
			songs[_current_song].length % 60);
		strcat(songinfo,tmp);
	}

	sprintf(tmp," <%d>", zoltan_time_left());
	strcat(songinfo,tmp);

	return(songinfo);
}


/**
 * zoltan_poll - Polls the CD
 *
 * Polls the sound system. If file is a CDROM, tests if external player
 * has dead (by ending the song or aborting); in that case,
 * move to next song and play it. On CD Audio, asks the
 * interface what track is currently playing.
 * This function must be called frequently.
 */
void zoltan_poll(void)
{
	struct cdrom_subchnl cdsc;

	if(_current_song==-1)
	{
		if(_player_pid != -1)
			_zoltan_kill_player();

		return;
	}

	if(_media_type==TYPE_CDROM || _media_type==TYPE_HD)
	{
		if(_player_pid==-1)
		{
			zoltan_change_song(1);

			if(_current_song != -1)
				zoltan_play();
		}
	}
	else
	if(_media_type==TYPE_CDA)
	{
		/* ask CD drive what track is being played */
		memset(&cdsc,'\0',sizeof(struct cdrom_subchnl));
		cdsc.cdsc_format=CDROM_MSF;

		if(ioctl(_cd_fd, CDROMSUBCHNL, &cdsc) < 0)
			return;

		if(cdsc.cdsc_audiostatus == 0x11 ||
		   cdsc.cdsc_audiostatus == 0x12)
		{
			if(cdsc.cdsc_trk - 1 != _current_song)
			{
				_song_startup_time=(int)time(NULL);

				if(songs[cdsc.cdsc_trk - 1].skip || _zoltan_shuffle)
				{
					zoltan_change_song(1);

					if(_current_song == -1)
						zoltan_stop();
					else
						zoltan_play();
				}
				else
					_current_song=cdsc.cdsc_trk-1;
			}
		}
		else
			_current_song=-1;
	}

	zoltan_notify();

	zoltan_save_song_cache();
}


/**
 * zoltan_cddb_seek_cache - Seeks the local cddb cache
 *
 * Seeks the local cddb cache for a file with the same name
 * as the CDDB id contained in _cddb_id.
 * If that file is found, the information is loaded as song info
 * and 1 is returned.
 */
int zoltan_cddb_seek_cache(void)
{
	FILE * f;
	char * ptr;

	snprintf(buffer,sizeof(buffer),"%s/%08x",_cddb_cache,_cddb_id);

	if((f=fopen(buffer,"r"))==NULL)
		return(0);

	while((fgets(buffer,sizeof(buffer),f))!=NULL)
	{
		/* chop */
		if(buffer[strlen(buffer)-1]=='\n')
			buffer[strlen(buffer)-1]='\0';
		if(buffer[strlen(buffer)-1]=='\r')
			buffer[strlen(buffer)-1]='\0';

		if(strncmp(buffer,"DTITLE=",7)==0)
			zoltan_add_group(buffer+7, 0);
		else
		if(strncmp(buffer,"TTITLE",6)==0)
		{
			ptr=buffer+6;
			while(*ptr!='\0' && *ptr!='=')
				ptr++;

			if(*ptr)
			{
				ptr++;
				zoltan_add_song(ptr,NULL,NULL,0,0);
			}
		}
	}

	fclose(f);

	return(1);
}


/**
 * zoltan_cddb_info - Fills CDDB info
 *
 * Seeks in the local cddb cache (via zoltan_cddb_seek_cache())
 * and, if not found, queries a remote server via zoltan_cddb_query().
 */
void zoltan_cddb_info(void)
{
	/* try to find it in the cache first */
	if(! zoltan_cddb_seek_cache())
	{
		/* ask the world about it */
		if(zoltan_cddb_query())
		{
			if(! zoltan_cddb_seek_cache())
				_num_songs=_cda_num_tracks;
		}
		else
			_num_songs=_cda_num_tracks;
	}
}


/**
 * zoltan_mount - 'Mounts' the CD
 *
 * 'Mounts' the CD. If a CDROM, really mount it into _mount_point;
 * If a CDA, reads info about it using zoltan_cda_info().
 * Returns zero if the CD could not be 'mounted'.
 */
int zoltan_mount(void)
{
	int x;

	_num_songs=0;
	_num_groups=0;

	zoltan_stop();
	_current_song=-1;

	/* exit if no CD */
	if(!zoltan_cd_test())
	{
		_media_type=TYPE_UNKNOWN;
		return(0);
	}

	/* umount current CDROM, if any */
	snprintf(buffer, sizeof(buffer), "umount %s", _mount_point);
	system(buffer);

	/* try first to mount it as a CDROM */
	snprintf(buffer, sizeof(buffer), "mount %s", _mount_point);
	x=system(buffer);

	if(!x)
	{
		/* CDROM was mounted without error */
		_media_type=TYPE_CDROM;

		/* travel the directory tree, searching for songs */
		zoltan_song_find(_mount_point, _mount_point, 0);
	}
	else
	{
		/* CD must be CDA */
		_media_type=TYPE_CDA;

		/* gets physical info about the CD */
		zoltan_cda_info();

		/* gets readable info */
		zoltan_cddb_info();
	}

	/* moves to beginning */
	_current_song=-1;

	/* it must not be paused anymore */
	_zoltan_paused=0;

	zoltan_shuffle(0);

	return(1);
}


static int _song_cmp(const void * e1, const void * e2)
{
	struct song * s1;
	struct song * s2;

	s1=(struct song *)e1;
	s2=(struct song *)e2;

	/* order first by group */
	if(s1->group!=s2->group)
		return(s1->group - s2->group);

	/* if group is a playlist, order by songid */
	if(groups[s1->group].is_playlist)
		return(s1->songid - s2->songid);

	/* if on the same group, order by name */
	return(strcmp(s1->songinfo, s2->songinfo));
}


/**
 * zoltan_hd - Searches songs in the local hard disk directory
 *
 * Searches the local hard disk directory (pointed by _hd_song_dir)
 * for songs, and fills the database with anything found.
 */
void zoltan_hd(void)
{
	zoltan_stop();

	_num_songs=0;
	_num_groups=0;

	_media_type=TYPE_HD;

	zoltan_song_find(_hd_song_dir, _hd_song_dir, 0);

	/* sort the songs */
	qsort(songs, _num_songs, sizeof(struct song), _song_cmp);

	_current_song=-1;

	/* it must not be paused anymore */
	_zoltan_paused=0;

	zoltan_shuffle(0);
}


/**
 * zoltan_umount - 'Unmounts' the CD
 *
 * 'Unmounts' the CD. If it's a CDROM, it really unmounts it.
 * If it's a CD Audio, stop it. Anyway, it's ejected.
 */
void zoltan_umount(void)
{
	if(_media_type==TYPE_CDROM)
	{
		_zoltan_kill_player();

		snprintf(buffer, sizeof(buffer), "umount %s", _mount_point);
		system(buffer);
	}
	else
		ioctl(_cd_fd, CDROMSTOP);

	ioctl(_cd_fd, CDROMEJECT);

	/* now we dont know what kind of CD is in, if any */
	_media_type=TYPE_UNKNOWN;
}


/**
 * zoltan_add_player - Adds a player
 * @ext: the file extension the player is able to play
 * @path: the path to the player program
 *
 * Adds a player to the internal database.
 */
void zoltan_add_player(char * ext, char * path)
{
	struct player * p;

	p=&players[_num_players];

	if(p->ext!=NULL) free(p->ext);
	if(p->path!=NULL) free(p->path);

	p->ext=(char *)malloc(strlen(ext)+1);
	p->path=(char *)malloc(strlen(path)+1);

	strcpy(p->ext,ext);
	strcpy(p->path,path);

	_num_players++;
}


static unsigned short _zoltan_crc_1(unsigned short crc, int c)
{
	int x, n;

	x=((int)crc << 8) | c;

	for(n=0;n < 8;n++)
	{
		x<<=1;

		if(x & 0x01000000)
			x^=0x01102100;
	}

	return((x & 0x00ffff00) >> 8);
}


unsigned short _zoltan_crc(char * string)
{
	unsigned short crc;

	for(crc=0;*string;string++)
		crc=_zoltan_crc_1(crc, *string);

	crc=_zoltan_crc_1(crc,'\0');
	crc=_zoltan_crc_1(crc,'\0');

	return(crc);
}


/**
 * zoltan_cache_store_song - Caches the length of current song
 *
 * Stores the length of the current song in the cache, and marks
 * it as dirty for later saving.
 */
void zoltan_cache_store_song(void)
{
	unsigned short crc;

	if(songs[_current_song].length==0 ||
	   songs[_current_song].filename==NULL) return;

	/* calculate crc of the song name */
	crc=_zoltan_crc(songs[_current_song].filename);

	if(crc==0) return;

	/* store */
	song_cache[(int)crc]=songs[_current_song].length;

	_song_cache_updates++;
}


/**
 * zoltan_search_cache - Searches the cache for the length of a song
 * @filename: Filename of the song
 *
 * Searches in the cache the length of the song which @filename
 * is sent as argument. If song length is not cached, returns 0.
 */
int zoltan_search_cache(char * filename)
{
	unsigned short crc;

	/* calculate crc of the song name */
	crc=_zoltan_crc(filename);

	return(song_cache[(int)crc]);
}


/**
 * zoltan_load_song_cache - Loads the song cache
 *
 * Loads the song length cache.
 */
void zoltan_load_song_cache(void)
{
	FILE * f;
	int n,i;
	unsigned short s;
	int format;

	/* cleanup */
	memset(song_cache,'\0',sizeof(song_cache));

	if((f=fopen(_song_cache_file,"r"))==NULL)
		return;

	format=fgetc(f);

	if(format==0) /* song cache format 0: ZoltanPlayer 0.2.5 */
	{
		/* { unsigned short crc, int length } for each entry */

		for(n=0;n < SONG_CACHE_SIZE;n++)
		{
			if(fread(&s,sizeof(unsigned short),1,f)==0)
				break;

			if(fread(&i,sizeof(int),1,f)==0)
				break;

			song_cache[(int)s]=i;
		}
	}
	else
	if(format==1) /* song cache format 1: ZoltanPlayer 0.2.6 */
	{
		/* { short length } for each entry */

		for(n=0;n < SONG_CACHE_SIZE;n++)
		{
			if(fread(&s,sizeof(unsigned short),1,f)==0)
				break;

			song_cache[n]=(int) s;
		}
	}

	fclose(f);

	_song_cache_updates=0;
}


/**
 * zoltan_save_song_cache - Saves the song cache
 *
 * Saves the song length cache, if marked as dirty.
 */
void zoltan_save_song_cache(void)
{
	FILE * f;
	int n;
	short s;

	if(! _song_cache_updates) return;

	if((f=fopen(_song_cache_file,"w"))==NULL)
		return;

	fputc(SONG_CACHE_FORMAT,f);

	for(n=0;n < SONG_CACHE_SIZE;n++)
	{
		s=(short)song_cache[n];
		fwrite(&s,sizeof(short),1,f);
	}

	fclose(f);

	_song_cache_updates=0;
}


static int _song_cmp_random(const void * e1, const void * e2)
{
	return((rand() % 3)-1);
}


/**
 * zoltan_shuffle - Shuffles all songs
 * @toggle: shuffle yes / no
 *
 * If @toggle is 1, shuffles all the songs; they will be
 * played in random order. A value of 0, resets normal order.
 */
void zoltan_shuffle(int toggle)
{
	int n;

	srand(time(NULL));

	_zoltan_shuffle=toggle;

	for(n=0;n < _num_songs;n++) _song_order[n]=n;

	if(toggle)
		qsort(_song_order, _num_songs,
			sizeof(int), _song_cmp_random);
}


/**
 * zoltan_load_config - Loads the configuration file
 * @configfile: the file name
 *
 * Loads the configuration file. Aborts if file can't be open.
 */
void zoltan_load_config(char * configfile)
{
	FILE * f;
	char * ptr;
	char * ptr2;

	if((f=fopen(configfile,"r"))==NULL)
	{
		printf("Can't open config file '%s'\n",configfile);
		exit(1);
	}

	while((fgets(buffer,sizeof(buffer),f))!=NULL)
	{
		buffer[strlen(buffer)-1]='\0';

		if(buffer[0]=='\0' || buffer[0]=='#') continue;

		if((ptr=strchr(buffer, ':'))!=NULL)
		{
			*ptr='\0'; ptr++;
			while(*ptr==' ') ptr++;

			if(strcmp(buffer,"cd_device")==0)
				strncpy(_cd_device,ptr,sizeof(_cd_device));
			else
			if(strcmp(buffer,"mixer_device")==0)
				strncpy(_mixer_device,ptr,sizeof(_mixer_device));
			else
			if(strcmp(buffer,"mount_point")==0)
				strncpy(_mount_point,ptr,sizeof(_mount_point));
			else
			if(strcmp(buffer,"cddb_cache")==0)
				strncpy(_cddb_cache,ptr,sizeof(_cddb_cache));
			else
			if(strcmp(buffer,"hd_song_dir")==0)
				strncpy(_hd_song_dir,ptr,sizeof(_hd_song_dir));
			else
			if(strcmp(buffer,"cddb_host")==0)
				strncpy(_cddb_host,ptr,sizeof(_cddb_host));
			else
			if(strcmp(buffer,"cddb_cgi")==0)
				strncpy(_cddb_cgi,ptr,sizeof(_cddb_cgi));
			else
			if(strcmp(buffer,"port")==0)
				_zoltan_port=atoi(ptr);
			else
			if(strcmp(buffer,"player")==0)
			{
				if((ptr2=strchr(ptr,' '))!=NULL)
				{
					*ptr2='\0'; ptr2++;
					while(*ptr2==' ') ptr2++;

					zoltan_add_player(ptr,ptr2);
				}
			}
			else
			if(strcmp(buffer,"startup_volume")==0)
				_zoltan_volume=atoi(ptr);
			else
			if(strcmp(buffer,"volume_increment")==0)
				_zoltan_volume_increment=atoi(ptr);
			else
			if(strcmp(buffer,"notify_command")==0)
				strncpy(_notify_command,ptr,sizeof(_notify_command));
			else
			if(strcmp(buffer,"song_cache_file")==0)
				strncpy(_song_cache_file,ptr,sizeof(_song_cache_file));
		}
	}

	fclose(f);
}


/**
 * zoltan_set_volume - Sets the current global volume
 *
 * Sets the current volume (stored in the variable _zoltan_volume)
 * using the system mixer. Global, PCM and CD volumes are currently
 * set at the same time to the same value (this will probably change).
 */
void zoltan_set_volume(void)
{
	int v;

	if(_mixer_fd != -1)
	{
		v=(_zoltan_volume << 8) | _zoltan_volume;
		ioctl(_mixer_fd, SOUND_MIXER_WRITE_VOLUME, &v);

		v=(_zoltan_volume << 8) | _zoltan_volume;
		ioctl(_mixer_fd, SOUND_MIXER_WRITE_PCM, &v);

		v=(_zoltan_volume << 8) | _zoltan_volume;
		ioctl(_mixer_fd, SOUND_MIXER_WRITE_CD, &v);
	}
}


/**
 * zoltan_time_left - Returns the time until the end of the song
 *
 * Returns the time (in seconds) left until the end of the song.
 * If no song length is known, it returns an approximated time.
 * Used mainly to set the reload page time.
 */
int zoltan_time_left(void)
{
	int t;

	/* by default, a probable value */
	t=150;

	/* do we know the song length? */
	if(songs[_current_song].length)
	{
		t=songs[_current_song].length -
			((int)time(NULL) - _song_startup_time);
	}

	if(t < 0)
		t=0;

	/* add a second */
	return(t + 1);
}


/**
 * zoltan_startup - Core startup
 * @configfile: the configuration file
 *
 * Opens the devices and clears the database.
 * Returns 0 if configuration file could not be open (fatal error).
 */
int zoltan_startup(char * configfile)
{
	/* clear database */
	memset(songs,'\0',sizeof(songs));
	memset(groups,'\0',sizeof(groups));
	memset(players,'\0',sizeof(players));
	_num_songs=0;
	_num_groups=0;
	_num_players=0;
	_current_song=-1;

	zoltan_load_config(configfile);

	zoltan_load_song_cache();

	/* open cd device */
	if((_cd_fd=open(_cd_device, O_RDONLY | O_NONBLOCK))==-1)
		return(0);

	/* open mixer device */
	if(_mixer_device[0]!='\0')
		_mixer_fd=open(_mixer_device, O_RDWR);

	return(1);
}
