/*

    zoltan_http.c - ZoltanPlayer CD/MP3 Player

    HTTP communication code.

    Copyright (C) 2001 Angel Ortega <angel@triptico.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

    http://www.triptico.com

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>

#include "zoltan.h"

/*******************
	Data
********************/

/* the cddb hostname */
char _cddb_host[512]="www.freedb.org";

/* the cgi */
char _cddb_cgi[512]="/~cddb/cddb.cgi";

/* communications buffer */
static char comm_buffer[4096];


/*******************
	Code
********************/

/**
 * zoltan_write - Write a string into a socket
 * @a: the socket
 * @str: the string
 *
 * Writes a string into a socket, calculating it size.
 */
void zoltan_write(int a, char * str)
{
	send(a,str,strlen(str),0);
}


/**
 * zoltan_writef - Write a string into a socket, with formatting
 * @a: the socket
 * @fmt: the format string
 *
 * Writes a string into a socket, with the same formatting
 * as the printf() series of functions.
 */
void zoltan_writef(int a, char * fmt, ...)
{
	va_list argptr;

	va_start(argptr, fmt);
	vsnprintf(comm_buffer, sizeof(comm_buffer), fmt, argptr);
	va_end(argptr);

	zoltan_write(a,comm_buffer);
}


/**
 * zoltan_client_socket - Opens a connection to a remote server
 * @server: the remote host name
 * @port: the remote port
 *
 * Opens a connection to a remote server. Very generic; can be
 * used for any TCP/IP connection.
 * Returns the connected socket, or -1 on error.
 */
int zoltan_client_socket(char * server, int port)
{
	int s;
	struct hostent * h;
	struct sockaddr_in sa;

	/* find host */
	if((h=gethostbyname(server))==NULL)
		return(-1);

	memcpy((char *)&sa.sin_addr,(char *)h->h_addr,h->h_length);
	sa.sin_family=h->h_addrtype;
	sa.sin_port=htons(port);

	if((s=socket(AF_INET, SOCK_STREAM, 0))==-1)
		return(-1);

	if(connect(s,(struct sockaddr *)&sa,sizeof(struct sockaddr))==-1)
		return(-1);

	return(s);
}


/**
 * zoltan_send_http_query - Sends an HTTP query
 * @s: the socket
 * @query: the query
 * @host: the host name
 *
 * Sends an HTTP query with a HTTP 1.0 GET command. The
 * HTTP response from the server will be waiting in the socket.
 */
void zoltan_send_http_query(int s, char * query, char * host)
{
	zoltan_write(s,"GET ");
	zoltan_write(s,query);
	zoltan_write(s," HTTP/1.0\r\n");
	zoltan_write(s,"Host: ");
	zoltan_write(s,host);
	zoltan_write(s,"\r\n\r\n");

	/* HTTP response is now waiting in the socket */
}


/**
 * zoltan_drop_http_response - Drops the HTTP response from the server
 * @s: the socket
 *
 * Drops the HTTP response from the server, as it's not used
 * in Zoltan.
 */
void zoltan_drop_http_response(int s)
{
	unsigned char c,lc;

	c='\0';
	for(;;)
	{
		if(recv(s,&c,1,0)<=0) break;
		if(c=='\r') continue;

		if(c=='\n' && lc=='\n')
			break;
		lc=c;
	}
}


/**
 * _zoltan_cddb_query - Makes a partial query
 * @genre: the possible genre of the current CD
 *
 * Makes a query in the specified genre for the CDDB id
 * contained in _cddb_id.
 * Returns 1 if an entry in the CDDB database could be found.
 */
static int _zoltan_cddb_query(char * genre)
{
	int s,r;
	unsigned char c;
	char tmp[16];
	FILE * f;
	int ret;

	s=zoltan_client_socket(_cddb_host, 80);

	snprintf(comm_buffer,sizeof(comm_buffer),
		"%s?cmd=cddb+read+%s+%08x&hello=me+where.i.am+Zoltan+%s&proto=1",
		_cddb_cgi, genre, _cddb_id, VERSION);

	printf("%s\n",comm_buffer);

	zoltan_send_http_query(s,comm_buffer,_cddb_host);

	zoltan_drop_http_response(s);

	/* read three char response */
	recv(s,tmp,3,0);
	r=atoi(tmp);

	if(r==210)
	{
		/* entry found! */
		/* drop rest of line */
		while(recv(s,&c,1,0)>0)
		{
			if(c=='\n') break;
		}

		/* write it to database */
		sprintf(comm_buffer,"%s/%08x",_cddb_cache,_cddb_id);

		if((f=fopen(comm_buffer,"w"))!=NULL)
		{
			while(recv(s,&c,1,0)>0)
				putc(c,f);

			fclose(f);
		}

		ret=1;
	}
	else
		ret=0;

	close(s);

	return(ret);
}


/**
 * zoltan_cddb_query - Makes a complete CDDB query
 *
 * Makes a complete CDDB query for the CDDB id contained in
 * _cddb_id, trying all different genres.
 * Returns 1 if the entry could be found.
 */
int zoltan_cddb_query(void)
{
	/* the cddb genres, rearranged for my own pleasure */
	char * genres[]={ "rock", "misc", "newage", "classical",
			  "folk", "jazz", "blues", "soundtrack",
			  "reggae", "country", "data",
			  NULL };
	int n;

	for(n=0;genres[n]!=NULL;n++)
	{
		printf("Testing '%s'...\n",genres[n]);
		if(_zoltan_cddb_query(genres[n]))
		{
			printf("OK!\n");
			return(1);
		}
		printf("No.\n");
	}

	zoltan_add_group("No Info", 0);
	for(n=0;n < _cda_num_tracks;n++)
		zoltan_add_song("No Info",NULL,NULL,0,0);

	return(0);
}


/**
 * zoltan_server_socket - Opens a TCP/IP server
 * @port: the server port
 *
 * Opens a TCP/IP server on the specified port. The returned socket
 * is marked as non-blocking. Very generic; can be used for any
 * TCP/IP server.
 * Returns the rendez-vous socket.
 */
int zoltan_server_socket(int port)
{
	int s,n;
	struct sockaddr_in server;

	memset(&server,'\0',sizeof(server));
	server.sin_family=AF_INET;
	server.sin_port=htons(port);

	if((s=socket(AF_INET,SOCK_STREAM,0))==-1)
		return(-1);

	/* reuse address */
	n=1;
	setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&n,sizeof(n));

	if(bind(s,(struct sockaddr *)&server,sizeof(server))==-1)
		return(-2);

	listen(s,5);

	/* set to non-blocking */
	fcntl(s, F_SETFL, O_NONBLOCK);

	return(s);
}


/**
 * zoltan_accept - Accepts remote connections
 * @s: the rendez-vous socket
 *
 * Accepts remote connections from the specified rendez-vous socket.
 * Returns -1 on error, 0 if the calling would block (i.e., there is
 * no connection waiting) or the connection socket otherwise.
 */
int zoltan_accept(int s)
{
	int c,size;
	struct sockaddr_in client;

	size=sizeof(struct sockaddr_in);

	/* try accept */
	if((c=accept(s,(struct sockaddr *)&client,&size))==-1)
	{
		if(errno==EWOULDBLOCK)
			return(0);
		else
			return(-1);
	}

	return(c);
}


/**
 * zoltan_client - Sends a command to a zoltan server
 * @server: the zoltan server hostname
 * @port: the zoltan server port
 * @command: the command to be sent
 *
 * Sends a command to a zoltan server. The server reply
 * is stored in reply_buffer, if defined.
 * Returns the size of the data read, or 0 if the
 * connection could not be done.
 */
int zoltan_client(char * server, int port, char * command,
	char * reply_buffer, int reply_size)
{
	int s,r;

	/* connect */
	if((s=zoltan_client_socket(server, port))==-1)
		return(0);

	zoltan_send_http_query(s, command, server);

	zoltan_drop_http_response(s);

	/* read the rest as reply */
	r=recv(s,reply_buffer,reply_size,0);

	close(s);

	return(r);
}
